﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Text.RegularExpressions;

namespace LSystems
{
    internal class ProductionRule
    {
        public List<Type> newLeft = new List<Type>();
        public List<Type> left = new List<Type>();
        public List<Type> strict = new List<Type>();
        public List<Type> right = new List<Type>();
        public List<Type> newRight = new List<Type>();
        public MethodInfo method;

        private static Regex productionRegex = new Regex(@"^([a-z,A-Z,0-9,\ ]*<{2})?([a-z,A-Z,0-9,\ ]*<)?([a-z,A-Z,0-9,\ ]+)(>[a-z,A-Z,0-9,\ ]*)?(>{2}[a-z,A-Z,0-9,\ ]*)?$");

        public static ProductionRule GenerateRule(MethodInfo method)
        {
            ProductionAttribute production = 
                (ProductionAttribute)Attribute.GetCustomAttribute(method, typeof(ProductionAttribute), false);

            if (production == null)
            {
                return null;
            }

            Match match = productionRegex.Match(production.Schema);
            if (!match.Success)
            {
                throw new InvalidOperationException("Production string is not correct (Regex match failure).");
            }

            if (match.Groups.Count != 6)
            {
                throw new InvalidOperationException("Internal error while parsing production.");
            }

            var newLeft = FindNames(match.Groups[1].Value, "<<");
            var left = FindNames(match.Groups[2].Value, "<");
            var strict = FindNames(match.Groups[3].Value, String.Empty);
            var right = FindNames(match.Groups[4].Value, ">");
            var newRight = FindNames(match.Groups[5].Value, ">>");

            var allItems = newLeft.Concat(left).Concat(strict).Concat(right).Concat(newRight).ToArray();
            var methodParametersNames = (from ParameterInfo info in method.GetParameters() select info.Name).ToArray();

            if (!allItems.SequenceEqual(methodParametersNames))
            {
                throw new InvalidOperationException("List of parameters in production do not match to list of method parameters.");
            }

            var parametersList = method.GetParameters().ToList();

            return new ProductionRule()
            {
                method = method,
                newLeft = (from string l in newLeft select parametersList.Find(p => p.Name == l).ParameterType).ToList(),
                left = (from string l in left select parametersList.Find(p => p.Name == l).ParameterType).ToList(),
                strict = (from string l in strict select parametersList.Find(p => p.Name == l).ParameterType).ToList(),
                right = (from string l in right select parametersList.Find(p => p.Name == l).ParameterType).ToList(),
                newRight = (from string l in newRight select parametersList.Find(p => p.Name == l).ParameterType).ToList()
            };
        }

        private static string[] FindNames(string s, string excludeName)
        {
            return
                (from string name in s.Split(new[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries)
                 where name != excludeName
                 select name).ToArray();
        }


        public object TryApply(
            List<object> currentString, 
            int currentIndex, 
            List<object> newLeftString, 
            List<object> newRightString,
            object systemDefinition)
        {
            // Check new left part.
            if (newLeftString != null)
            {
                int newLeftStart = newLeftString.Count - this.newLeft.Count;
                if (newLeftStart < 0)
                {
                    // New left is too short.
                    return null;

                }

                if (!this.newLeft.SequenceEqual(
                        newLeftString.GetTypesSequence(newLeftStart, this.newLeft.Count)))
                {
                    // New left does not match.
                    return null;
                }
            }

            // Check new left part.
            if (newRightString != null)
            {
                if (newRightString.Count < this.newRight.Count)
                {
                    // New right is too short.
                    return null;

                }

                if (!this.newRight.SequenceEqual(
                        newRightString.GetTypesSequence(0, this.newRight.Count)))
                {
                    // New right does not match.
                    return null;
                }
            }

            // Check left part.
            if (currentIndex < this.left.Count ||
                !this.left.SequenceEqual(currentString.GetTypesSequence(
                    currentIndex - this.left.Count, this.left.Count)))
            {
                return null;
            }

            // Check strict predecessor.
            if (currentIndex + this.strict.Count > currentString.Count ||
                !this.strict.SequenceEqual(currentString.GetTypesSequence(
                currentIndex, this.strict.Count)))
            {
                return null;
            }

            // Check right part.
            if (currentIndex + this.strict.Count + this.right.Count > currentString.Count ||
                !this.right.SequenceEqual(
                    currentString.GetTypesSequence(currentIndex + this.strict.Count, this.right.Count)))
            {
                return null;
            }

            // Rule is OK, create parameters list.
            List<object> newLeftParams = 
                (newLeftString == null) 
                ? newLeftString.GetSequence(newLeftString.Count - this.newLeft.Count, this.newLeft.Count).AsList()
                : new List<object>();

            List<object> middleParams =  currentString.GetSequence(
                    currentIndex - this.left.Count, 
                    this.left.Count + this.strict.Count + this.right.Count).AsList();

            List<object> newRightParams = 
                (newRightString == null) 
                ? newRightString.GetSequence(0, this.newRight.Count).AsList()
                : new List<object>();

            return this.method.Invoke(
                systemDefinition,
                newLeftParams.Concat(middleParams).Concat(newRightParams).ToArray());
        }
    }
}
